import LRU from "lru-cache";
import type { Query, UpObject } from "./index";
import { UpListing } from "./index";
import type {
  Address,
  ADDRESS_TYPE,
  AttributeListingResult,
  EntityListing,
  IJob,
  IValue,
  ListingResult,
  PutInput,
  PutResult,
  StoreInfo,
  VaultInfo,
} from "./types";
import type { AddressComponents, UpEndWasmExtensions } from "./wasm";
import debug from "debug";
import { browser } from "./util";

const dbg = debug("upend:api");

export type { AddressComponents };

export class UpEndApi {
  private instanceUrl = "";
  private readonly wasmExtensions: UpEndWasmExtensions | undefined = undefined;
  public readonly timeout: number;

  private queryOnceLRU = new LRU<string, UpListing>({ max: 128 });
  private inFlightRequests: { [key: string]: Promise<UpListing> | null } = {};

  constructor(config: {
    instanceUrl?: string;
    wasmExtensions?: UpEndWasmExtensions;
    timeout?: number;
  }) {
    this.setInstanceUrl(config.instanceUrl || "http://localhost:8093");
    this.wasmExtensions = config.wasmExtensions;
    this.timeout = config.timeout || 30_000;
  }

  public setInstanceUrl(apiUrl: string) {
    this.instanceUrl = apiUrl.replace(/\/+$/g, "");
  }

  public get apiUrl() {
    return this.instanceUrl + "/api";
  }

  public async fetchEntity(
    address: string,
    options?: ApiFetchOptions,
  ): Promise<UpObject> {
    dbg("Fetching Entity %s", address);
    const signal = this.getAbortSignal(options);
    const entityFetch = await fetch(`${this.apiUrl}/obj/${address}`, {
      signal,
    });
    const entityResult = (await entityFetch.json()) as EntityListing;
    const entityListing = new UpListing(entityResult.entries);
    return entityListing.getObject(address);
  }

  public async fetchEntry(address: string, options?: ApiFetchOptions) {
    dbg("Fetching entry %s", address);
    const signal = this.getAbortSignal(options);
    const response = await fetch(`${this.apiUrl}/raw/${address}`, { signal });
    const data = await response.json();
    const listing = new UpListing({ address: data });
    return listing.entries[0];
  }

  public async query(
    query: string | Query,
    options?: ApiFetchOptions,
  ): Promise<UpListing> {
    const queryStr = query.toString();

    const cacheResult = this.queryOnceLRU.get(queryStr);
    if (!cacheResult) {
      if (!this.inFlightRequests[queryStr]) {
        dbg(`Querying: ${query}`);
        this.inFlightRequests[queryStr] = new Promise((resolve, reject) => {
          const signal = this.getAbortSignal(options);
          fetch(`${this.apiUrl}/query`, {
            method: "POST",
            body: queryStr,
            keepalive: true,
            signal,
          })
            .then(async (response) => {
              if (!response.ok) {
                reject(
                  `Query ${queryStr} failed: ${response.status} ${
                    response.statusText
                  }: ${await response.text()}}`,
                );
              }
              resolve(new UpListing(await response.json()));
              this.inFlightRequests[queryStr] = null;
            })
            .catch((err) => reject(err));
        });
      } else {
        dbg(`Chaining request for ${queryStr}...`);
      }
      return await (this.inFlightRequests[queryStr] as Promise<UpListing>); // TODO?
    } else {
      dbg(`Returning cached: ${queryStr}`);
      return cacheResult;
    }
  }

  public async putEntry(
    input: PutInput,
    options?: ApiFetchOptions,
  ): Promise<PutResult> {
    dbg("Putting %O", input);
    const signal = this.getAbortSignal(options);
    const response = await fetch(`${this.apiUrl}/obj`, {
      method: "PUT",
      headers: { "Content-Type": "application/json" },
      body: JSON.stringify(input),
      signal,
    });

    return await response.json();
  }

  public async putEntityAttribute(
    entity: Address,
    attribute: string,
    value: IValue,
    provenance?: string,
    options?: ApiFetchOptions,
  ): Promise<Address> {
    dbg("Putting %s = %o for %s (%s)", attribute, value, entity, provenance);
    let url = `${this.apiUrl}/obj/${entity}/${attribute}`;
    if (provenance) {
      url += `?provenance=${provenance}`;
    }

    const signal = this.getAbortSignal(options);
    const response = await fetch(url, {
      method: "PUT",
      headers: { "Content-Type": "application/json" },
      body: JSON.stringify(value),
      signal,
    });

    return await response.json();
  }

  public async putBlob(
    fileOrUrl: File | URL,
    options?: ApiFetchOptions & { onProgress?: (ev: ProgressEvent) => void },
  ): Promise<Address> {
    dbg("Putting Blob: %O", fileOrUrl);

    const formData = new FormData();
    if (fileOrUrl instanceof File) {
      formData.append(fileOrUrl.name, fileOrUrl);
    } else {
      formData.append("@url", fileOrUrl.toString());
    }
    const signal = this.getAbortSignal(options);

    if (browser && fileOrUrl instanceof File) {
      dbg("Using XHR for file upload");
      const xhrdbg = debug("upend:api:xhr");
      const xhr = new XMLHttpRequest();
      signal.addEventListener("abort", () => xhr.abort());
      for (const event of [
        "loadstart",
        "load",
        "loadend",
        "progress",
        "abort",
        "error",
      ] as const) {
        xhr.addEventListener(event, (ev) => xhrdbg(`XHR ${event}: %O`, ev));
        xhr.upload.addEventListener(event, (ev) =>
          xhrdbg(`XHR upload ${event}: %O`, ev),
        );
        if (options?.onProgress) {
          xhr.upload.addEventListener(event, options.onProgress);
        }
      }
      return new Promise((resolve, reject) => {
        xhr.open("PUT", `${this.apiUrl}/blob`, true);
        xhr.onload = () => {
          if (xhr.status >= 200 && xhr.status < 300) {
            try {
              resolve(JSON.parse(xhr.responseText));
            } catch (e) {
              reject(e);
            }
          } else {
            reject(xhr.statusText);
          }
        };
        xhr.send(formData);
      });
    } else {
      const response = await fetch(`${this.apiUrl}/blob`, {
        method: "PUT",
        body: formData,
        signal,
      });

      if (!response.ok) {
        throw Error(await response.text());
      }

      return await response.json();
    }
  }

  public async deleteEntry(
    address: Address,
    options?: ApiFetchOptions,
  ): Promise<void> {
    dbg("Deleting entry %s", address);
    const signal = this.getAbortSignal(options);
    await fetch(`${this.apiUrl}/obj/${address}`, { method: "DELETE", signal });
  }

  public getRaw(address: Address, preview = false) {
    return `${this.apiUrl}/${preview ? "thumb" : "raw"}/${address}`;
  }

  public async fetchRaw(
    address: Address,
    preview = false,
    options?: ApiFetchOptions,
  ) {
    dbg("Getting %s raw (preview = %s)", address, preview);
    const signal = this.getAbortSignal(options);
    return await fetch(this.getRaw(address, preview), { signal });
  }

  public async refreshVault(options?: ApiFetchOptions) {
    dbg("Triggering vault refresh");
    const signal = this.getAbortSignal(options);
    return await fetch(`${this.apiUrl}/refresh`, { method: "POST", signal });
  }

  public async nativeOpen(address: Address, options?: ApiFetchOptions) {
    dbg("Opening %s natively", address);
    const signal = this.getAbortSignal(options);
    return fetch(`${this.apiUrl}/raw/${address}?native=1`, { signal });
  }

  public async fetchRoots(options?: ApiFetchOptions): Promise<ListingResult> {
    dbg("Fetching hierarchical roots...");
    const signal = this.getAbortSignal(options);
    const response = await fetch(`${this.apiUrl}/hier_roots`, { signal });
    const roots = await response.json();
    dbg("Hierarchical roots: %O", roots);
    return roots;
  }

  public async fetchJobs(options?: ApiFetchOptions): Promise<IJob[]> {
    // dbg("Fetching jobs...");
    const signal = this.getAbortSignal(options);
    const response = await fetch(`${this.apiUrl}/jobs`, { signal });
    return await response.json();
  }

  public async fetchAllAttributes(
    options?: ApiFetchOptions,
  ): Promise<AttributeListingResult> {
    dbg("Fetching all attributes...");
    const signal = this.getAbortSignal(options);
    const response = await fetch(`${this.apiUrl}/all/attributes`, { signal });
    const result = await response.json();
    dbg("All attributes: %O", result);
    return await result;
  }

  public async fetchInfo(options?: ApiFetchOptions): Promise<VaultInfo> {
    dbg("Fetching vault info...");
    const signal = this.getAbortSignal(options);
    const response = await fetch(`${this.apiUrl}/info`, { signal });
    const result = await response.json();
    dbg("Vault info: %O", result);
    return result;
  }

  public async fetchStoreInfo(
    options?: ApiFetchOptions,
  ): Promise<{ [key: string]: StoreInfo }> {
    dbg("Fetching store info...");
    const signal = this.getAbortSignal(options);
    const response = await fetch(`${this.apiUrl}/stats/store`, { signal });
    const result = await response.json();
    dbg("Store info: %O");
    return await result;
  }

  public async getAddress(
    input: { urlContent: string } | ADDRESS_TYPE,
    options?: ApiFetchOptions,
  ): Promise<string> {
    let response: Response;
    if (typeof input === "string") {
      if (this.wasmExtensions) {
        await this.wasmExtensions.init();
        return this.wasmExtensions.AddressTypeConstants[input];
      }
      const signal = this.getAbortSignal(options);
      response = await fetch(`${this.apiUrl}/address?type=${input}`, {
        signal,
      });
    } else {
      if ("urlContent" in input) {
        const signal = this.getAbortSignal(options);
        response = await fetch(
          `${this.apiUrl}/address?url_content=${input.urlContent}`,
          { signal },
        );
      } else {
        throw new Error("Input cannot be empty.");
      }
    }
    const result = await response.json();
    dbg("Address for %o = %s", input, result);
    return result;
  }

  public async addressToComponents(
    address: string,
  ): Promise<AddressComponents> {
    if (!this.wasmExtensions) {
      throw new Error("WASM extensions not supplied.");
    }
    await this.wasmExtensions.init();
    return this.wasmExtensions.addr_to_components(address);
  }

  public async componentsToAddress(
    components: AddressComponents,
  ): Promise<string> {
    if (!this.wasmExtensions) {
      throw new Error("WASM extensions not initialized.");
    }
    await this.wasmExtensions.init();
    return this.wasmExtensions.components_to_addr(components);
  }

  public async getVaultOptions(
    options?: ApiFetchOptions,
  ): Promise<VaultOptions> {
    const signal = this.getAbortSignal(options);
    const response = await fetch(`${this.apiUrl}/options`, { signal });
    return await response.json();
  }

  public async setVaultOptions(
    options: VaultOptions,
    apiOptions?: ApiFetchOptions,
  ): Promise<void> {
    const payload: Record<string, unknown> = {};

    if (options.blob_mode) {
      const blob_mode: Record<string, unknown> = {};
      blob_mode[options.blob_mode] = null;
      payload["blob_mode"] = blob_mode;
    }

    const signal = this.getAbortSignal(apiOptions);
    const response = await fetch(`${this.apiUrl}/options`, {
      method: "PUT",
      headers: { "Content-Type": "application/json" },
      body: JSON.stringify(payload),
      signal,
    });

    if (!response.ok) {
      throw Error(await response.text());
    }
  }

  private getAbortSignal(options: ApiFetchOptions | undefined) {
    const controller = options?.abortController || new AbortController();
    const timeout = options?.timeout || this.timeout;
    if (timeout > 0) {
      setTimeout(() => controller.abort(), timeout);
    }
    return controller.signal;
  }
}

export interface ApiFetchOptions {
  timeout?: number;
  abortController?: AbortController;
}

export type VaultBlobMode = "Flat" | "Mirror" | "Incoming";
export interface VaultOptions {
  blob_mode: VaultBlobMode;
}
